# Copyright (c) 2008,2009,2010 Abraham Heifets
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import tkSimpleDialog
import tkMessageBox
from pymol import cmd
from pymol import models
import pymol
import sys, random, math
import munkres 
import re, os
import keynat
from datetime import datetime

colors = [
"aquamarine",
#"black",
"blue",
"bluewhite",
"br0",
"br1",
"br2",
"br3",
"br4",
"br5",
"br6",
"br7",
"br8",
"br9",
"brightorange",
"brown",
"carbon",
"chartreuse",
"chocolate",
"cyan",
"darksalmon",
"dash",
"deepblue",
"deepolive",
"deeppurple",
"deepsalmon",
"deepsalmon",
"deepteal",
"density",
"dirtyviolet",
"firebrick",
"forest",
"gray",
"green",
"greencyan",
"grey",
"hotpink",
"hydrogen",
"lightblue",
"lightmagenta",
"lightorange",
"lightpink",
"lightteal",
"lime",
"limegreen",
"limon",
"magenta",
"marine",
"nitrogen",
"olive",
"orange",
"oxygen",
"palecyan",
"palegreen",
"paleyellow",
"pink",
"purple",
"purpleblue",
"raspberry",
"red",
"ruby",
"salmon",
"sand",
"skyblue",
"slate",
"smudge",
"splitpea",
"sulfur",
"teal",
"tv_blue",
"tv_green",
"tv_orange",
"tv_red",
"tv_yellow",
"violet",
"violetpurple",
"warmpink",
"wheat",
"white",
"yellow",
"yelloworange"
]

# MacPyMOL default install still uses pre-version 2.5 python.
def all(iterable):
    for element in iterable:
        if not element:
            return False
    return True

def any(iterable):
    for element in iterable:
        if element:
            return True
    return False

def intersection(x,y):
    return [a for a in x if a in y]

def foldl(f, i, l):
    for x in l:
        i = f(i, x)
    return i

def sort_f(l, f):
    seq = [(f(x), x) for x in l]
    seq.sort()
    return [x for (junk, x) in seq]

def all_combos(list_of_lists):
    if len(list_of_lists) == 0:
        return [[]]
    else:
        ret_val = []
        for a in list_of_lists[0]:
            for b in all_combos(list_of_lists[1:]):
                b.append(a)
                ret_val.append(b)
        return ret_val

def print_nested(l, indent=""):
    if isinstance(l, list): # specifically list because I don't want to break up tuples (since I use them as keys, a fair bit).
        print indent,
        print l
        for x in l:
            print_nested(x, indent+".")
    else:
        print indent,
        print l

def init_defaults():
    if not pymol.stored.__dict__.has_key("ligalign_settings"):
        pymol.stored.ligalign_settings = {}    
    pymol.stored.ligalign_settings["cutoff"] = 0.1
    pymol.stored.ligalign_settings["max_splits"] = None
    pymol.stored.ligalign_settings["min_fragment_size"] = 6
    pymol.stored.ligalign_settings["fingerprint_scoring"] = (lambda (a,b): arbitrary_scoring(a,b))
    pymol.stored.ligalign_settings["distant_residue_distance"] = 4.5
init_defaults()

def ligalign_set(*args):
    if not pymol.stored.__dict__.has_key("ligalign_settings"):
        init_defaults()

    if not args:
        for setting in pymol.stored.ligalign_settings.items():
            print "%s = %s", setting
    else:
        if len(args) == 2:
            # unwrap into setting and value, then set
            pymol.stored.ligalign_settings[args[0]] = float(args[1]) # pymol.stored.ligalign_settings[args[1]]
            pass
        else:
            print "I didn't quite understand that."
            raise QuietException
    pass
cmd.extend("ligalign_set", ligalign_set)

def random_color():
    return random.choice(colors)

def selection_from_atomlist(atomlist):
    return " or ".join(["(%s and id %d)" % (selection_name, id) for (selection_name, id) in atomlist])
def atomlist_from_selection(selection):
    return cmd.identify(selection,1)

def clean_sel(sel):
    cmd.alter(sel, "name='C1'")
    cmd.alter(sel, "resi='1'")
    cmd.alter(sel, "segi=''")
    cmd.alter(sel, "chain=''")
    cmd.alter(sel, "resn=''")
    cmd.alter(sel, "b=0.0")
    cmd.alter(sel, "q=0.0")

def rmsd_on_models(m1, m2, mapping, group):
    if len(group) == 0:
        return 0.0
    else:
        atoms1 = []
        atoms2 = []
        for index in group:
            # translate index to id
            id = -1
            for a in m1.atom:
                if a.index == index:
                    id = a.id
                    if id in mapping:
                        atoms1.append(a)
                        break
            for a in m2.atom:
                if id in mapping and a.id == mapping[id]:
                    atoms2.append(a)
                    break
        
        coords1 = [a.coord for a in atoms1]
        coords2 = [a.coord for a in atoms2]
        try: 
            (t1, t2, rot_mtx, rmsd) = models.fit(coords1, coords2)  # Note: this is MODELS.fit(), *not* CMD.fit()
        except Exception:  # TODO: figure out what's causing the errors, internal to the PyMOL code
            rmsd = 9999.0
        return rmsd

def align_by_ligand((baseline, junk_mapping), to_align): # selection, list of selections
    if not pymol.stored.__dict__.has_key("ligand_mappings"):
        pymol.stored.ligand_mappings = {}    
    maps = [(baseline, junk_mapping)]
    for (x,mapping) in to_align:
        if not mapping:
            mapping = build_mapping(x, baseline) #pymol.stored.ligand_mappings.setdefault((x,baseline), build_mapping(x, baseline))
        maps.append((x,mapping))
        # convert to pair list
        pairs = []
        x_model = cmd.get_model(x)
        for atom in x_model.atom:
            if atom.id in mapping:
                k,v = atom.id , mapping[atom.id]
                pairs.append("(%s and id %s)" % (x, k))                
                pairs.append("(%s and id %s)" % (baseline, v))
        cmd.pair_fit(*pairs)
    for (name, mapping) in maps:
        pymol.stored.ligand_mappings[name] = (baseline, mapping)    # cache the mappings for possible future display
    return maps

def build_protein_specific_active_sites(distance=4):
    psas = []
    if False:
        # TODO: FIXME: BUG: Right now, I'm using a hacky correspondence between protein name and ligand name.  It would be better to build real datastructures for this.
        for p in pymol.stored.protein_names:
            # assume that proteins with a _%d suffix are fragments.  Depending on assumptions like this is undesirable.
            pname = re.match("(.*)_\d+$", p)
            if pname:
                prot = pname.group(1)
                if prot in pymol.stored.protein_names:
                    cleaned_protein_name = prot
                else:
                    cleaned_protein_name = p
            else:
                cleaned_protein_name = p
            psas.append("(("+p+" within %d" % distance +" of "+ p +"_ligand) and not "+cleaned_protein_name+"_ligand)")           
    psas = ["(("+p+" within %d" % distance +" of "+ p +"_ligand) and not (byres "+p+"_ligand))" for p in pymol.stored.protein_names]
#    psas = ["(("+p+" within %d" % distance +" of "+ p +"_ligand) and not "+p+"_ligand)" for p in cleaned_protein_names]
    return psas

def build_ligand_union(ligand_selects):
    ligand_union = "ligand_union"
    cmd.select( "ligand_union", "((" + " or ".join(["(%s)" % l for l in ligand_selects]) + "))")
    cmd.deselect()
    return ligand_union

def build_active_site(ligand_selects, distance=5):
    # build ligand-union selection
    # then build AS selection
    ligand_union = "(" +" or ".join(["(%s)" % l for l in ligand_selects])+ ")"
    AS_select = "((all within %d"  % distance +" of "+ligand_union+") and not "+ligand_union+")"
    AS_select =  "(" +" or ".join(["(%s)" % l for l in build_protein_specific_active_sites(distance)])+ ")"
    cmd.select("as_select", AS_select)
    cmd.deselect()
    return AS_select

def draw_active_site(AS_select):
    cmd.show("spheres", AS_select)
    cmd.zoom(AS_select)

def viz(setting="1", ligands="ligand_union", active_site="as_select", full_residues=True, entire_protein=False, small_balls=False, distance=4):
    # global view effects
#    cmd.hide("lines", "all")
#    cmd.hide("everything", "all and resn HOH")
    if full_residues=="True" or full_residues=="true":
        tmp_selection_name = active_site
        full_active_site = "full_residues"
        cmd.select(full_active_site, "(byres "+active_site+")")
        cmd.deselect()
        active_site = full_active_site
    if entire_protein=="True" or entire_protein=="true":
        for p in pymol.stored.protein_names:
            cmd.show("lines", p)
        cmd.zoom()
    if small_balls=="True" or small_balls=="true":
        # viz 2) ligand-sticks, AS-little balls
        cmd.show("spheres", active_site)
        cmd.show("lines", active_site)
        cmd.set("sphere_scale", 0.25, active_site)
        pass

    if setting == "1":
        # clear
        cmd.hide("everything", "all")
        # viz 1) ligand-sticks, AS-balls
        cmd.show("sticks", ligands)
        cmd.show("spheres", active_site)
        cmd.set("sphere_scale", 1.0, active_site)
        pass
    elif setting == "2":
        # clear
        cmd.hide("everything", "all")
        # viz 2) ligand-sticks, AS-little balls
        cmd.show("sticks", ligands)
        cmd.show("spheres", active_site)
        cmd.show("lines", active_site)
        cmd.set("sphere_scale", 0.25, active_site)
        pass
    elif setting == "3":
        # clear
        cmd.hide("everything", "all")
        # viz 3) ligand-sticks, as-closed surface transparent
        cmd.create("as_surface", active_site)
        cmd.show("sticks", ligands)
        cmd.show("spheres", "as_surface")
        cmd.show("lines", "as_surface")
        cmd.set("sphere_scale", 0.25, "as_surface")
        cmd.show("surface", "as_surface")
        cmd.set("transparency", 0.4)
        pass
    elif setting == "4":
        # clear
        cmd.hide("everything", "all")
        # viz 4) ligand-sticks, as-closed surface opaque
        cmd.create("as_surface", active_site)
        cmd.show("sticks", ligands)
        cmd.show("spheres", "as_surface")
        cmd.show("lines", "as_surface")
        cmd.set("sphere_scale", 0.25, "as_surface")
        cmd.show("surface", "as_surface")
        cmd.set("transparency", 0.0)
        pass
    elif setting == "5":
        # clear
        cmd.hide("everything", "all")
        # viz 5) ligand-sticks, as-open surface
        cmd.show("sticks", ligands)
        cmd.show("surface", active_site)
        cmd.set("transparency", 0.0)
        pass
    elif setting == "6":
        # clear
        cmd.hide("everything", "all")
        # viz 6) ligand-balls, as-sticks
        cmd.show("spheres", ligands)
        cmd.show("sticks", active_site)
        cmd.set("sphere_scale", 1.0, active_site)
        pass
    elif setting == "7":
        # clear
        cmd.hide("everything", "all")
        # viz 8)
        cmd.show("sticks", ligands)
        psas = build_protein_specific_active_sites()
        for i in range(len(pymol.stored.protein_names)):
            p = psas[i]
            if full_residues == "True":
                p = "(byres "+p+")"
            p_name = pymol.stored.protein_names[i]
            l_name = pymol.stored.ligand_names[i]
            prot_surface = p_name + "_surface"
            cmd.create(prot_surface, p)
            cmd.show("surface", prot_surface)
            cmd.hide("spheres", prot_surface)
            color = random_color()
            print "coloring",l_name,color
            cmd.color(color, prot_surface)
            cmd.color(color, l_name)
        cmd.set("transparency", 0.4)
        pass
    elif setting == "8":
        # clear
        cmd.hide("everything", "all")
        # viz 9)
        cmd.show("sticks", ligands)
        psas = build_protein_specific_active_sites()
        for i in range(len(pymol.stored.protein_names)):
            p = psas[i]
            if full_residues == "True":
                p = "(byres "+p+")"
            p_name = pymol.stored.protein_names[i]
            l_name = pymol.stored.ligand_names[i]
            prot_surface = p_name + "_surface"
            cmd.create(prot_surface, p)
            cmd.show("surface", prot_surface)
            cmd.show("sticks",  prot_surface)
            cmd.set("stick_radius", 0.10, prot_surface)
            cmd.hide("spheres", prot_surface)
            color = random_color()
            print "coloring",l_name,color
            cmd.color(color, prot_surface)
            cmd.color(color, l_name)
            cmd.color(color, p_name)
#            cmd.show("lines", p_name)
        cmd.set("transparency", 0.4)
        cmd.set("two_sided_lighting", 1)
        pass
    elif setting == "9":
        # clear
        cmd.hide("everything", "all")
        # viz 9)
        cmd.show("sticks", ligands)
        psas = build_protein_specific_active_sites()
        for i in range(len(pymol.stored.protein_names)):
            p = psas[i]
            if full_residues == "True":
                p = "(byres "+p+")"
            p_name = pymol.stored.protein_names[i]
            l_name = pymol.stored.ligand_names[i]
            prot_surface = p_name + "_surface"
            cmd.create(prot_surface, p)
            cmd.show("sticks",  prot_surface)
            cmd.set("stick_radius", 0.10, prot_surface)
            cmd.hide("spheres", prot_surface)
            color = random_color()
            print "coloring",l_name,color
            cmd.color(color, prot_surface)
            cmd.color(color, l_name)
            cmd.color(color, p_name)
#            cmd.show("lines", p_name)
        cmd.set("transparency", 0.4)
        cmd.set("two_sided_lighting", 1)
        pass
    elif setting == "10":
        # clear
        cmd.hide("everything", "all")
        # viz 9)
        cmd.show("sticks", ligands)
        psas = build_protein_specific_active_sites()
        for i in range(len(pymol.stored.protein_names)):
            p = psas[i]
            if full_residues == "True":
                p = "(byres "+p+")"
            p_name = pymol.stored.protein_names[i]
            l_name = pymol.stored.ligand_names[i]
            prot_surface = p_name + "_surface"
            cmd.create(prot_surface, p)
            cmd.show("spheres", prot_surface)
            cmd.set("sphere_scale", 0.05, prot_surface)
            color = random_color()
            print "coloring",l_name,color
            cmd.color(color, prot_surface)
            cmd.color(color, l_name)
            cmd.color(color, p_name)
        pass
    elif setting == "11":
        # clear
        cmd.hide("everything", "all")
        # viz 8)
        cmd.show("sticks", ligands)
        cmd.set("transparency", 0.4)
        pass
    else:
        cmd.hide("everything", "all")
        pass
    cmd.center(ligands)
    cmd.zoom(active_site)
cmd.extend("ligalign_viz", viz)


def draw_mapping(baseline, to_align, mapping):
        if not mapping:
            mapping = build_mapping(to_align, baseline)
        x_model = cmd.get_model(to_align)
        name = "mapping_%s_to_%s" % (to_align, baseline)
        for atom in x_model.atom:
            if atom.id in mapping:
                k,v = atom.id , mapping[atom.id]
                cmd.distance(name, "(%s and id %s)" % (to_align, k), "(%s and id %s)" % (baseline, v))
                cmd.show("dashes", name)
                cmd.hide("labels", name)
#                print name, ":", "(%s and id %s)" % (to_align, k), "(%s and id %s)" % (baseline, v)
def ligalign_show_mapping(sel):
    if not pymol.stored.__dict__.has_key("ligand_mappings"):
        return # no mappings means 
    if sel not in pymol.stored.ligand_mappings:
        ligand = ("%s_ligand" % sel)
        if ligand in pymol.stored.ligand_mappings:
            sel = ligand
        else:
            print "I don't have any mapping for",sel
            return
    (baseline,mapping) = pymol.stored.ligand_mappings[sel]
    draw_mapping(baseline, sel, mapping)
cmd.extend("ligalign_show_mapping", ligalign_show_mapping)


class Centroid:
    def __init__(self, x,y,z):
        self.x = x
        self.y = y
        self.z = z
    def distance(self, c):
        return ((self.x-c.x)**2+(self.y-c.y)**2+(self.z-c.z)**2)**0.5

def compute_centroids(protein):
    model = cmd.get_model(protein)
    # for each residue in active_site, figure out its atoms
    # protein / segi / chain / resi define a residue uniquely
    residue_atoms = {}
    for atom in model.atom:
        (res_type, ra) = residue_atoms.setdefault((atom.segi, atom.chain, atom.resi), (atom.resn, []))
        ra.append(atom)
        # for each residue, compute centroid/pseudocenter
    avg = lambda (index, atom_list): sum([atom.coord[index] for atom in atom_list])/float(len(atom_list))
    centroids = dict([(residue, [avg((i, atom_list)) for i in [0,1,2]]) for (residue, (res_name, atom_list)) in residue_atoms.items()])
    res_names={}
    for (residue, (res_name, centroid)) in residue_atoms.items():
        res_names[residue] = res_name
    return residue_atoms, res_names, centroids


def find_distances(baseline_centroids, centroids):
    # for each centroid in active_site, figure out closest neighbor in baseline
    distances = {}
    for (residue, centroid) in centroids.items():
        dists = [(((centroid[0]-b_c[0])**2+(centroid[1]-b_c[1])**2+(centroid[2]-b_c[2])**2) ** 0.5, b_r) for (b_r, b_c) in baseline_centroids.items()]
        dists.sort()
        distances[residue] = dists
    # returns residue * [(float,residue)] dictionary
    return distances

def compute_distance(centroid, b_c):
    return centroid.distance(b_c)

def find_clusters_by_distances_squared(baseline_centroids, centroids, range, filter_f):
    # for each centroid in active_site, figure out closest neighbor in baseline
    distances = {}
    inverse_distances = {}
    for (residue, centroid) in centroids.items():
        dists = []
        for (b_r, b_c) in baseline_centroids.items():
            d=compute_distance(centroid, b_c) #((centroid[0]-b_c[0])**2+(centroid[1]-b_c[1])**2+(centroid[2]-b_c[2])**2)**0.5
            if d <= range and filter_f((residue, b_r)):
                dists.append((d,b_r))
                inverse_distances.setdefault(b_r, []).append((d,residue))
#        dists = [(((centroid[0]-b_c[0])**2+(centroid[1]-b_c[1])**2+(centroid[2]-b_c[2])**2), b_r) for (b_r, b_c) in baseline_centroids.items()]
        dists.sort()
        distances[residue] = dists
    for d in inverse_distances.values():
        d.sort()
    # returns residue * [(float,residue)] dictionary
    return distances, inverse_distances


def residue_stddevs(baseline, *proteins_tuple):
    baseline_model = cmd.get_model(baseline)
    baseline_atoms, baseline_res_names, baseline_centroids = compute_centroids(baseline)
    proteins=list(proteins_tuple)
    distances = {}#dict([(residue, [centroid_distances])])
    for protein in proteins:
        print protein
        # Compute all centroids of all residues in all proteins.
        residue_atoms, res_names, protein_centroids = compute_centroids(protein)
        # for each residue in the baseline active site,
        # find minimum distance to some centroid in each other protein.
        dists = find_distances(protein_centroids,baseline_centroids) # b/c find_distances returns residues from 2nd argument, pass in baseline 2nd.
        for (residue, (res_name,atoms)) in baseline_atoms.items():
            (distance, nearest) = dists[residue][0]
            res_dist_list = distances.setdefault(residue, [])
            res_dist_list.append(distance)
    
    print distances
    # compute stddev
    avg = lambda (dist_list): sum(dist_list)/float(len(dist_list))
    residues = baseline_atoms.keys()
    residues.sort()
    stddevs = [(r, sum([(d - avg(distances[r]))**2 for d in distances[r]])**0.5) for r in residues]
    
    for (r, stddev) in stddevs:
        print r,"=",stddev

    print "Average standard deviation:"
    if stddevs:
        print sum([stddev for (r,stddev) in stddevs])/len(stddevs)
    else:
        print "N/A"
cmd.extend("ligalign_residue_stddevs", residue_stddevs)

# Compaction is necessary because PyMOL crashes on long selection strings
class ResidueCompactor:
    class ResRange:
        def __init__(self, min, max):
            self.min = min
            self.max = max
        def __str__(self):
            if self.min == self.max:
                return "(resi %s)"%self.min
            else:
                return "(resi %s-%s)"%(self.min,self.max)
                
    def compact(self, r):
        residues = [self.ResRange(x,x) for x in r]
        l = len(residues)
        modified = True
        while modified:
            modified = False
            for x in residues:
                merged = False
                for y in residues:
                    if x.max == y.min-1:
                        y.min = x.min
                        merged = True
                if merged:
                   residues.remove(x)
                   modified = True
                   break
        return [str(x) for x in residues]

    def __init__(self, protein, r):
        # Residues are expected to be (atom.segi, atom.chain, atom.resi)
        # segregate by segi and chains, then compact
        # TODO: are segi and alt handled consistently throughout the code?  Replace (s,c,r) representation of residues w/ residue class.
        res = {}
        for (segi, chain, resi) in r:
            (all_residues, nonnumeral_residues) = res.setdefault((segi,chain), ([],[]))
            # Check for weird residues in a couple of ways.

            # Check that residue *can* be compacted without errors!
            # Example: [select (1hlp and (segi '') and (chain A) and (resi 53-53))] and [select (1hlp and (segi '') and (chain A) and (resi 53))] yield different results!!
            select = "(%s" % protein
            if len(segi) > 0:
                select += (" and (segi %s)"  % segi)
            else:
                select += (" and (segi '')")
            if len(chain) > 0: 
                select += (" and (chain %s)"  % chain)
            else:
                select += (" and (chain '')")
            cmd.select("ligalign_internal_test", "%s and (resi %s))" % (select,resi))
            nAtoms1 = cmd.count_atoms("ligalign_internal_test")
            cmd.select("ligalign_internal_test", "%s and (resi %s-%s))" % (select,resi,resi))
            nAtoms2 = cmd.count_atoms("ligalign_internal_test")

            if (nAtoms1 == nAtoms2) and resi.isdigit():
                all_residues.append(int(resi))
            else:
                # apparently there are residue ids that contain letters.
                nonnumeral_residues.append(resi)
        cmd.delete("ligalign_internal_test")
        selects = []
#        print "res = ", res
        for ((segi, chain), (reses, weird_reses)) in res.items():
            reses.sort() # for easier reading later
            str = []
            if len(segi) > 0:
                str.append("(segi %s)"  % segi)
            else:
                str.append("(segi '')")

            if len(chain) > 0: 
                str.append("(chain %s)" % chain)
            else:
                str.append("(chain '')")

            res_selection_string = "("
            if len(reses) > 0:
                res_selection_string += "("+" or ".join(self.compact(reses))+")"
            else:
                res_selection_string += "(none)"
            res_selection_string += " or "
            if len(weird_reses) > 0:
                res_selection_string += "("+" or ".join(["(resi %s)"%x for x in weird_reses])+")"
            else:
                res_selection_string += "(none)"
            res_selection_string += ")"
            str.append(res_selection_string)

            selects.append("("+" and ".join(str)+")")

        if len(selects) > 0:
            self.str = "("+" or ".join(selects)+")"
        else:
            self.str = "none"

    def __str__(self):
        return self.str


def cluster_diameter(pairwise_distances, centroids_by_protein, cluster):
    try:
        diameters = []
        for (p1,r1) in cluster:
            for (p2,r2) in cluster:
                if ((p1,r1),(p2,r2)) not in pairwise_distances:
                    # compute it!
                    centroid = centroids_by_protein[p1][r1]
                    b_c      = centroids_by_protein[p2][r2]
                    d=compute_distance(centroid, b_c) #((centroid[0]-b_c[0])**2+(centroid[1]-b_c[1])**2+(centroid[2]-b_c[2])**2)**0.5
                    pairwise_distances[((p1,r1),(p2,r2))] = d
                diameters.append(pairwise_distances[((p1,r1),(p2,r2))])
#        return max([pairwise_distances[(x,y)]**0.5 for x in cluster for y in cluster])
        return max(diameters)
    except KeyError, e:
        print "e:", e, "cluster:",cluster
        raise KeyError
def maximal_intersection(rest, match_percent, max_length, pairwise_distances, maximum_diameter, centroids_by_protein, debug_flag = False, intersection_so_far=None, omittable_proteins_remaining=0):
#    omittable_proteins_remaining = (max_length - int(match_percent * max_length)) - (max_length - len(rest))
    omittable_proteins_remaining = len(rest) - int(match_percent * max_length)
    minimum_intersection_bound = float(match_percent * max_length)
    if len(rest) > max_length:
        print "ERROR:", max_length, "len(rest):", len(rest)
        for r in rest:
            print "--",r
        raise Error 
    if omittable_proteins_remaining >=0:
        foo =  maximal_intersection_inner(rest, minimum_intersection_bound, omittable_proteins_remaining, pairwise_distances, maximum_diameter, centroids_by_protein, debug_flag, None)
        return foo
    else: 
        return []

debug_max_count = 0
def maximal_intersection_inner(rest, minimum_intersection_bound, omittable_proteins_remaining, pairwise_distances, maximum_diameter, centroids_by_protein, debug_flag, intersection_so_far):
    global debug_max_count
    debug_max_count += 1
    if debug_flag: print "................int:", intersection_so_far, " rest:",rest
    too_small = lambda x: float(len(x)) < minimum_intersection_bound 
    if rest:
        if debug_flag: print "rest[0]oik3:", rest[0]
        best_intersection = []
        best_intersection_diameter = sys.maxint
        alternatives = sort_f(rest[0], len)
        for option in alternatives:
            if intersection_so_far:
                new_int = intersection(intersection_so_far, option)
            else:
                new_int = option
            if too_small(new_int):
                continue  # Using this alternative causes failure already.  don't bother refining the intersection.
            else:
                #  Why can't we do this pruning?
#                diameter = cluster_diameter(pairwise_distances, centroids_by_protein, new_int)
#                if diameter > best_intersection_diameter or diameter > maximum_diameter:
#                    continue # we've found a tighter cluster already, so don't bother extending this cluster
                maximal_int = maximal_intersection_inner(rest[1:], minimum_intersection_bound, omittable_proteins_remaining, pairwise_distances, maximum_diameter, centroids_by_protein, debug_flag, new_int)
                if too_small(maximal_int):
                    continue # try the next alternative
                else:
                    # group the residues by proteins
                    prot_res = {}
                    for (p,r) in maximal_int:
                        prot_res.setdefault(p, []).append((p,r))
                    if too_small(prot_res.keys()): # check that we have enough proteins to complete a cluster
                        continue
                    # now build possible maximal intersections by taking one residue from each protein
                    possible_ints = all_combos(prot_res.values())
                    if debug_max_count % 10000 == 0:
                        print "DEBUG: debug_max_count:",debug_max_count," len(possible_ints):",len(possible_ints)
                    for m_i in possible_ints:
                        if too_small(m_i):
                            continue
                        else:
                            # we found some intersection that is big enough to matter, now ensure smallest diameter
                            diameter = cluster_diameter(pairwise_distances, centroids_by_protein, m_i)
                            if diameter < best_intersection_diameter and diameter < maximum_diameter and len(best_intersection) <= len(m_i):
                                best_intersection_diameter = diameter
                                best_intersection = m_i
        if best_intersection:
            return best_intersection
        if omittable_proteins_remaining > 0:
            # if we aren't required to put all proteins in the intersection, then we can leave some out.  
            # So, since the current alternative options all overconstrain the intersection, try leaving this one out.
#            if intersection_so_far and len(intersection_so_far)<5:
#                print "rest:",rest,"ommitable:",omittable_proteins_remaining,"intersection_so_far",intersection_so_far
            return maximal_intersection_inner(rest[1:], minimum_intersection_bound, omittable_proteins_remaining-1, pairwise_distances, maximum_diameter, centroids_by_protein, debug_flag, intersection_so_far)
        else:
            return [] # none of the possible intersections was big enough!
    else:
        if debug_flag: print "intersection_so_far:", intersection_so_far
        if intersection_so_far:
#            if intersection_so_far and len(intersection_so_far)<5:
#                print "rest:",rest,"ommitable:",omittable_proteins_remaining,"intersection_so_far",intersection_so_far
            return intersection_so_far
        else:
            return []
    
def get_residue_name(residue_name):
    residue_resn_abbreviations = {
        'ALA':'A', 
        'ARG':'R', 
        'ASN':'N', 
        'ASP':'D', 
        'CYS':'C', 
        'GLU':'E', 
        'GLN':'Q', 
        'GLY':'G', 
        'HIS':'H', 
        'ILE':'I', 
        'LEU':'L', 
        'LYS':'K', 
        'MET':'M', 
        'PHE':'F', 
        'PRO':'P', 
        'SER':'S', 
        'THR':'T', 
        'TRP':'W', 
        'TYR':'Y', 
        'VAL':'V', 
        }
    if residue_name in residue_resn_abbreviations:
        return residue_resn_abbreviations[residue_name]
    else:
        return "?"

# Categories taken from JC Nebel, "Generation of 3D templates of active sites of proteins with rigid prosthetic groups", Bioinformatics 2006
amino_acid_classes = {
    'ALA':['nonpolar'],
    'ARG':['basic'],  
    'ASN':['amidic'], 
    'ASP':['acidic'], 
    'CYS':['sulphide bond'], 
    'GLU':['acidic'], 
    'GLN':['amidic'], 
    'GLY':['nonpolar'], 
    'HIS':['basic'], 
    'ILE':['nonpolar'], 
    'LEU':['nonpolar'], 
    'LYS':['basic'], 
    'MET':['nonpolar'], 
    'PHE':['aromatic'], 
    'PRO':['nonpolar'], 
    'SER':['hydroxyl'], 
    'THR':['hydroxyl'], 
    'TRP':['aromatic'], 
    'TYR':['aromatic', 'hydroxyl'], 
    'VAL':['nonpolar'], 
    }
def get_amino_acid_class(residue_name):
    if residue_name in amino_acid_classes:
        return amino_acid_classes[residue_name]
    else:
        return residue_name

def color_by_amino_acid_class(selection):
    amino_acid_class_colors = {
        'acidic':'warmpink', #'ruby', #'firebrick', 
        'basic':'blue', 
        'amidic':'forest',#'green', 
        'sulphide bond':'yellow', 
        'nonpolar':'tv_orange', #'lightorange', #'brightorange', 
        'aromatic':'magenta', 
        'hydroxyl':'palecyan', 
        }
    for (amino_acid_type, amino_acid_class) in amino_acid_classes.items():
        cmd.color(amino_acid_class_colors[amino_acid_class[0]], "(%s and resn %s)"%(selection, amino_acid_type))
cmd.extend("ligalign_color_by_amino_acid_class", color_by_amino_acid_class)

def find_template(maximum_diameter=1.0, match_percent=1.0, filter_func="geometric", *proteins_tuple, **foo):
    print "Find template parameters: max_diameter=", maximum_diameter, " match_percent=", match_percent, " filter_func=", filter_func, " proteins=", proteins_tuple, " other=",foo
    # debug_test = lambda (p1,chain,resi):  (p1 == "1bu7" and int(resi) == 973)
    # debug_test = lambda (p1,chain,resi):  (p1 == "1w0g" and chain == "A" and int(resi) == 473) or (p1 == "1pq2" and chain == "A" and int(resi) == 467) or (p1 == "1bu7" and chain == "" and int(resi) == 63) or (p1 == "1h5z" and chain == "A" and int(resi) == 423)
    # debug_test = lambda (p1,chain,resi):  (p1 == "1pq2" and chain == "A" and int(resi) == 378) or (p1 == "1w0g" and chain == "A" and int(resi) == 385)
    # debug_test = lambda (p1,chain,resi):  (p1 == "1h5z" and chain == "A" and int(resi) == 431) or (p1 == "1w0g" and chain == "A" and int(resi) == 480)
    # debug_test = lambda (p1,chain,resi):  (p1 == "1h5z" and chain == "A" and int(resi) == 431) or (p1 == "1w0g" and chain == "A" and int(resi) == 480)
    # debug_test = lambda (p1,chain,resi):  (p1 == "1pq2" and chain == "A" and int(resi) == 85) or (p1 == "1w0g" and chain == "A" and int(resi) == 88)
    # debug_test = lambda (p1,chain,resi):  (p1 == "1ldm" and chain == "A" and resi == '80') or (p1 == "1ldn" and chain == "A" and resi == '81') or (p1 == "9ldt" and chain == "A" and resi == '81') or (p1 == "1hdx" and chain == "A" and resi == '241') or (p1 == "2ohx" and chain == "A" and resi == '241') or (p1 == "1gd1" and chain == "A" and resi == '73')
    debug_test = lambda (p1,chain,resi): False

    if filter_func == "geometric":
        filter_func = (lambda (residue1, residue2): True)
    elif filter_func == "chemical":
        # Ensure intersection is nonempty, rather than the result is equal, allows a residue to have multiple chemical classes.
        filter_func = lambda (residue1, residue2):  (residue1 == residue2) or (len(intersection(get_amino_acid_class(residue1), get_amino_acid_class(residue2))) > 0)
    elif filter_func == "identity":
        filter_func = lambda (residue1, residue2):  (residue1 == residue2) 
    else:
        # default to any clustering
        filter_func = (lambda (residue1, residue2): True)
    # we can probably speed this up (and, maybe, be more precise) by confining our investigation to the active site / (as_select) / build_active_site / build_protein_specific_active_sites
    # Perhaps we could use selection algebra to prune the number of residues for which we need to measure pairwise distances?
    maximum_diameter = float(maximum_diameter)
    match_percent = float(match_percent)
    proteins=list(proteins_tuple)
    clustered_proteins = []
    clustered_residues_by_protein = {}
    residue_names_by_protein = {}
    centroids_by_protein = {}
    all_extant_clusters = {}
    pairwise_distances = {}
    for p1 in proteins:
        print "Computing centroids and residue types for",p1
        residue_atoms, res_names, fooooooo = compute_centroids(p1)
        centroids = dict([(residue, Centroid(x[0], x[1], x[2])) for (residue,x) in fooooooo.items()])
        residue_names_by_protein[p1] = res_names
        centroids_by_protein[p1] = centroids
    cache_dists = {}
    print "Constructing adjacency matrix..."
    for p1 in proteins:
        clustered_residues = {}
#         proteins_without_p1.remove(p1) # don't compare protein against itself.  
        for p2 in proteins:
            print "...comparing",p1,"against",p2
            # for each residue in the baseline active site,
            # find minimum distance to some centroid in each other protein.  # TODO: perhaps restrict matching residues to be of similar chemical types, for speed
            if (p1,p2) in cache_dists:
                dists = cache_dists[(p1,p2)]
            else:
                 dists, inverse_dists = find_clusters_by_distances_squared(centroids_by_protein[p2], centroids_by_protein[p1], maximum_diameter, (lambda(a,b):filter_func((residue_names_by_protein[p1][a], residue_names_by_protein[p2][b]))))
                 cache_dists[(p1,p2)] = dists
                 cache_dists[(p2,p1)] = inverse_dists

            for (residue, close_list) in dists.items():
                (segi, chain, resi) = residue
                for (distance, closest_residue) in close_list:
                    if distance <= maximum_diameter and (filter_func((residue_names_by_protein[p1][residue], residue_names_by_protein[p2][closest_residue]))):
                        clustered_residues[residue] = clustered_residues.setdefault(residue, [])
                        clustered_residues[residue].append((p2, closest_residue))
                        # record pairwise distances
                        key = [(p1, residue), (p2, closest_residue)]
                        #key.sort()
                        pairwise_distances[tuple(key)] = distance
                        if debug_test((p1,chain,resi)):
                            print "Key:",key, "distance:",distance
        clustered_residues_by_protein[p1] = clustered_residues

    for (p1, clustered_residues) in clustered_residues_by_protein.items():
        print "Seeding maximal clique detector with %s residues."%p1
        filtered_clusters = {}
        for (residue, closest) in clustered_residues.items():
            (segi, chain, resi) = residue
            if debug_test((p1, chain, resi)):
                print p1, residue, "closest=", closest
                for (p3, y) in closest:
                    print "   protein:",p3, "residue:",y, "cluster:",clustered_residues_by_protein[p3][y]

            # Take the intersection of all cluster lists.  The result will be the shared residues, within the range) # guh... not if there's alternative residue options and we've got to choose between them.
            if True:
                cluster_alternatives = {}
                for (p2, x) in closest: # all nearby residues
                    their_nearby_residues = clustered_residues_by_protein[p2][x]
                    # organize alternative sets by protein...
                    alts = cluster_alternatives.setdefault(p2, [])
                    alts.append(their_nearby_residues)
                # now, we need one set of nearby residues, for every protein in closest.
                # Then, we find the maximal intersection and, if it's bigger than the match_percent, we use it.
                # Of course maximal intersection is NP-complete, no?
                alts = cluster_alternatives.values()
                size = 0
                inter = []
                if alts:
                    inter = maximal_intersection(alts, match_percent, len(proteins), pairwise_distances, maximum_diameter, centroids_by_protein);
                if inter:
                    filtered_clusters[residue] = inter#size
                    # make a canonical and hashable tuple from the intersection
                    key = list(inter)
                    key.sort()
                    all_extant_clusters[tuple(key)] = key
                    if debug_test((p1, chain, resi)):
                        print "... for p1=", p1, "residue=",residue, " found key=",key," from alts=",alts

        for (residue, inter) in filtered_clusters.items():
                (segi, chain, resi) = residue
                counts = len(inter)
                if debug_test((p1, chain, resi)):
                    print p1, residue, counts, (counts / float(len(proteins))),  (float(counts / float(len(proteins))) >= float(match_percent)), match_percent

    # I promise to filter out any clusters that are proper subsets of other clusters.
    # Let me do this filtering post-hoc, here, even though I suspect there's a more efficient way of doing this by integrating it in, above.
    maximalized_clusters = []
    for x in all_extant_clusters:
        if not any([tuple(intersection(x,y)) == tuple(x) for y in all_extant_clusters if x is not y]): # "is not" checks object equality
            maximalized_clusters.append(x)
    unique_clusters = maximalized_clusters #all_extant_clusters.keys()
    # re-define clustered_proteins to be a protein->(residue list) dict.
    prot_residue_mapping = {}
    for k in unique_clusters:
        for (p,r) in k:
            prot_residue_mapping.setdefault(p,[]).append(r)
    clustered_proteins = [(p,dict([(x,x) for x in r]).keys()) for (p,r) in prot_residue_mapping.items()] # remove duplicates



    print "\nFound %d distinct clusters, ordered by pairwise distance. " % len(unique_clusters),
    if len(unique_clusters) > 0:
        try:
            print "Average maximum pairwise distance:", (sum([cluster_diameter(pairwise_distances, centroids_by_protein, k) for k in unique_clusters])/len(unique_clusters))
        except KeyError, e:
            print "Error:", e
            print "(x,y):", (x,y)
            print "x:",x,"y:",y,"(x,y)inPD:",((x,y) in pairwise_distances)," ","(y,x)inPD:",((y,x) in pairwise_distances)
    else:
        print ""
    cluster_text = []
    # compute cluster columns
    unique_cluster_columns = []
    for k in unique_clusters:
        for (p,r) in k:
            if p not in unique_cluster_columns:
                unique_cluster_columns.append(p)
    for i in xrange(len(unique_clusters)):
        k = unique_clusters[i] 
        k_str = []
        k_by_protein = dict(k)
        for p in unique_cluster_columns:
            if p in k_by_protein:
                (s,c,r) = k_by_protein[p]
                x = "/%s/%s/%s/%s, "%(p,s,c,r)
            else:
                x = ""
            k_str.append(x.ljust(16))
        diameter = cluster_diameter(pairwise_distances, centroids_by_protein, k) #max([pairwise_distances[(x,y)]**0.5 for x in k for y in k])
        cluster_text.append((diameter, "".join(k_str)))
    cluster_text_justification = 0
    if cluster_text:
        cluster_text_justification = max([len(s) for (d,s) in cluster_text])
    cluster_text.sort()
    for i in xrange(len(cluster_text)):
        (diameter,k_str) = cluster_text[i]
        print ("Cluster %d:"%(i+1)).rjust(13),"  ", k_str.ljust(cluster_text_justification), "maximum pairwise distance:", diameter

    # Izhar's fantastic suggestion:  since PyMOL crashes on big selects, incrementally expand a select!
    cmd.select("template", "(none)") # initialize the template to nothing
    for (p1, residues) in clustered_proteins:
        for i in xrange(0, len(residues), 100): # take a clump of 100 residues and make the select
            clump = residues[i: i+100]
            select_string = "(template) or (%s and (%s))" % (p1, str(ResidueCompactor(p1, clump)))
            cmd.select("template", select_string) # or it together with the previously selected template
#    cmd.hide("sticks", "(template)")
    cmd.deselect()


    ########################################################################################################
    def build_distant_resi_set():
        if not pymol.stored.ligalign_settings.has_key("distant_residue_distance"):
            pymol.stored.ligalign_settings["distant_residue_distance"] = 4.5
        max_dist = float(pymol.stored.ligalign_settings["distant_residue_distance"])
        pymol.stored.ligalign_distant_resi_set = {}
        for (p,r) in clustered_proteins:
            pymol.stored.ligalign_distant_resi_set.setdefault(p,{})
            #                cmd.select("ligalign_junk2", "(%s and (template)) and not byres (%s and (resn NAD around 4.5) and (template))"%(p,p))
            #                cmd.select("ligalign_junk2", "(%s and not byres (%s and (resn NAD around 4.5)))"%(p[0:4],p[0:4])) from any NAD
            if cmd.count_atoms("%s_ligand"%(p,)) and cmd.count_atoms("(template)"):
#                cmd.select("ligalign_junk2", "(%s and (template) and not byres ((%s and %s_ligand) expand 4.5))"%(p,p,p)) # from THIS NAD
                cmd.select("ligalign_junk2", "(%s and (template) and not byres (%s_ligand expand %f))"%(p,p,max_dist)) # from THIS ligand
                cmd.iterate("ligalign_junk2", "stored.ligalign_distant_resi_set[\"%s\"][(alt,chain,resi)]=\"\\cellcolor[gray]{0.75}\""%p)
            res =  pymol.stored.ligalign_distant_resi_set[p].keys()
            res.sort(keynat.keynat_cmp)
            cmd.delete("ligalign_junk2")
    def print_latex_tables(ordered_clusters, caption=".", prune=False):
        build_distant_resi_set()
        low = 0
        step = 22 
        if prune:
            ordered_clusters = [d for d in ordered_clusters if not all([d[x] in pymol.stored.ligalign_distant_resi_set[x] for x in d if x in pymol.stored.ligalign_distant_resi_set])]
        num_tables = int(math.ceil(float(len(ordered_clusters)) / 22))  # 22 is about how many columns fit in one LaTeX table!  Make sure you're using {fullpage}
        step = int( math.ceil(float(len(ordered_clusters)) / max(num_tables,1)) ) # step is figuring out how many clusters to put per table, to get reasonably balanced tables
        while low < len(ordered_clusters):
            ordered_cluster_chunk = ordered_clusters[low:low+step]
            print "\\begin{table}[htp!]\n\\begin{center}\n\\tiny\n%\\scriptsize\n\\renewcommand{\\tabcolsep}{1pt}\n\\newcolumntype{R}{>{\\centering\\arraybackslash}X}"
            print "%\\begin{tabularx}{\\textwidth}{|c||",
            for d in ordered_cluster_chunk:
                if any([d[x] in pymol.stored.ligalign_distant_resi_set[x] for x in d if x in pymol.stored.ligalign_distant_resi_set]):
                    print ">{\\columncolor[gray]{0.75}}",
                print "R|", 
            print "}  \\hline"
            print "\\begin{tabularx}{\\textwidth}{|c||",
            for d in ordered_cluster_chunk:
                print "R|", 
            print "}  \\hline"
            print "PDB ID & ", " & ".join(["C%d"%i for i in range(low+1,low+1+min(step,len(ordered_cluster_chunk)))]), " \\\\ \\hline\\hline"
    # unique_clusters is ((p,(s,c,r)) list) list
            clusters_of_protein_residue_maps = [dict(k) for k in unique_clusters]
            for p in proteins:
                print p.upper().replace('\\','\\\\').replace('_','\_').replace('^','\^').replace('{','\{').replace('}','\}'), 
                for d in ordered_cluster_chunk:
                    print " & ",
                    if p in d: #and d[p] == reses[0]:
                #(a,c,r) = reses.pop(0)
                        (a,c,r) = d[p]
                        print "%s%s%s"%(pymol.stored.ligalign_distant_resi_set.get(p,{}).get((a,c,r),""), get_residue_name(residue_names_by_protein[p][(a,c,r)]), r), 
                        pass
                    else:
                        print "--", 
                print " \\\\"
                pass
            print "\\hline"
            print "Diameter",
            for d in ordered_cluster_chunk:
                print " & %.2f"%cluster_diameter(pairwise_distances, centroids_by_protein, [(p,d[p]) for p in d]),
            print " \\\\ \\hline"
            print "\\end{tabularx}\n\\caption{%s}\\label{.}\n\\end{center}\n\\end{table}\n"%caption
            low += step
    def order_clusters(unique_clusters):
        # ordered_clusters is a list of protein lists, where clusters are ordered by smallest (sequence-number) residue and proteins are unordered.
        ordered_clusters = []
        def cluster_cmp((a1,c1,r1),(a2,c2,r2)):
            return keynat.keynat_cmp(r1,r2)
        remaining_clusters = [dict(x) for x in unique_clusters]
        for p in proteins:
            # take all remaining clusters that have p in them
            clusters_to_insert = [x for x in remaining_clusters if p in x]
            clusters_to_insert.sort(lambda x,y: cluster_cmp(x[p], y[p]))
            remaining_clusters = [x for x in remaining_clusters if p not in x]
            # slot them into the ordered clusters appropriately, based on smallest p residue
            for c in clusters_to_insert:
                # find the last place we can insert c into, without disordering the list on inserting
                i = 0
                while i < len(ordered_clusters):
                    if p in ordered_clusters[i][0] and cluster_cmp(ordered_clusters[i][0][p], c[p]) > -1: #cluster_cmp(c[p], ordered_clusters[i][p]) > -1:
                        break
                    i += 1
                if i+1 < len(ordered_clusters) and  p in ordered_clusters[i+1][0] and cluster_cmp(ordered_clusters[i+1][0][p], c[p]) == 0:
                    ordered_clusters[i+1].append(c)
                else:
                    ordered_clusters.insert(i,[c])
        ordered_clusters_temp = []
        for x in ordered_clusters:
            ordered_clusters_temp.extend(x)
        ordered_clusters = ordered_clusters_temp
        return ordered_clusters

    print "Raw clusters LaTeX table:"
    ordered_clusters=order_clusters(unique_clusters)
    print_latex_tables(ordered_clusters, caption="Raw clusters")
#    print_latex_tables(ordered_clusters, caption="pruned merged", prune=True)


    print "Merged clusters LaTeX table:"
    merged_clusters = []
    for x in unique_clusters:
        merged = False
        for y in merged_clusters:
            if len(intersection(x,y)) > 0:
                w = dict(x)
                v = dict(y)
                if all([(a not in v or w[a] == v[a]) for a in w]):
                    for z in x:
                        if z not in y:
                            y.append(z)
                #y.extend(x)  # this will produce duplicates in the merged clusters but order_clusters will remove them.
                    merged = True
                #break
        if not merged:
            merged_clusters.append(list(x))

    ordered_clusters=order_clusters(merged_clusters)
#    print_latex_tables(ordered_clusters, caption="Merged clusters")
#    print_latex_tables(ordered_clusters, caption="Pruned merged clusters", prune=True)

    filtered_merged_clusters = []
    for x in merged_clusters:
        already_present = False
        for y in filtered_merged_clusters:
            if len(intersection(x,y)) > 0:
                already_present = True
                w = dict(x)
                x_diameter = cluster_diameter(pairwise_distances, centroids_by_protein, [(p,w[p]) for p in w])
                v = dict(y)
                y_diameter = cluster_diameter(pairwise_distances, centroids_by_protein, [(p,v[p]) for p in v])
                if len(x) > len(y) or (len(x) == len(y) and x_diameter < y_diameter):
                    filtered_merged_clusters.remove(y)
                    filtered_merged_clusters.append(x)
                    break
        if not already_present:
            filtered_merged_clusters.append(x)

    ordered_clusters=order_clusters(filtered_merged_clusters)
#    print_latex_tables(ordered_clusters, caption="filtered Merged clusters")

    print_latex_tables(ordered_clusters, caption="Merged clusters, within %.2f angstroms of the ligand."%(float(pymol.stored.ligalign_settings["distant_residue_distance"])), prune=True)

    build_distant_resi_set()
    ordered_clusters=order_clusters(filtered_merged_clusters)
    ordered_clusters = [d for d in ordered_clusters if not all([d[x] in pymol.stored.ligalign_distant_resi_set[x] for x in d if x in pymol.stored.ligalign_distant_resi_set])]
    cmd.select("close_template", "(none)")
    for p in proteins:
        for d in ordered_clusters:
            if p in d: #and d[p] == reses[0]:
                (a,c,r) = d[p]
                cmd.select("close_template", "close_template or /%s/%s/%s/%s"%(p,a,c,r))
#    cmd.hide("sticks","(close_template)")
    cmd.deselect()
    cmd.show("lines","close_template")

    for i in range(len(ordered_clusters)):
        d = ordered_clusters[i]
        name = i+1
        cmd.select("cluster_%d"%name, "(none)")
        for p in proteins:
            if p in d: #and d[p] == reses[0]:
                (a,c,r) = d[p]
                cmd.select("cluster_%d"%name, "cluster_%d or /%s/%s/%s/%s"%(name,p,a,c,r))
#        cmd.hide("sticks","(cluster_%d)"%name)
        cmd.deselect()

########################################################################################################
        

 #################### Generate output in ClustalW input format here
    print "ClustalW format:"
    bucket = {}
    for (p,res_list) in clustered_proteins:
        p1 = bucket.setdefault(p, {})
        for (s,c,r) in res_list:
            s1 = p1.setdefault(s, {})
            c1 = s1.setdefault(c, {})
            c1[(s,c,r)]=(s,c,r)
    for p in bucket: 
        for s in bucket[p]:
            for c in bucket[p][s]:
                model_selection = p
                if s and s != '':
                    model_selection += " and segi %s"  % s
                if c and c != '':
                    model_selection += " and chain %s" % c
                mod = cmd.get_model(model_selection)
                residue_tuples = mod.get_residues()

                # first, print unclustered residues
                clustalw_init = ""
                for res_index in range(len(residue_tuples)):
                    (start,stop) = residue_tuples[res_index]
                    letter =  get_residue_name(mod.atom[start].resn)
                    if letter == '?':
                        letter = 'X'
                    clustalw_init += letter
                    if ((res_index+1) % 75) == 0:
                        clustalw_init += "\n"
                print ">%s/%s/%s mol:protein length:%d LigAlign select:(%s)\n%s\n" % (p,s,c,(len(residue_tuples)+1),model_selection, clustalw_init)

                # now, print extracted cluster residues, for comparison
                res = bucket[p][s][c].keys()
                res.sort(keynat.keynat_cmp)
                clustered_res = dict([(r, get_residue_name(residue_names_by_protein[p][(s,c,r)])) for (s,c,r) in res])
                clustalw_init = ""
                for res_index in range(1,len(residue_tuples)+1): # I believe residues are 1-indexed
                    if str(res_index) in clustered_res:
                        letter =  clustered_res[str(res_index)]
                        if letter == '?':
                            letter = 'X'
                        clustalw_init += letter
                    else:
                        clustalw_init += "X"
                    if (res_index % 75) == 0:
                        clustalw_init += "\n"
                print ">%s/%s/%s_cluster mol:protein length:%d LigAlign select:(%s)\n%s\n" % (p,s,c,(len(residue_tuples)+1),model_selection, clustalw_init)
    print ""
 #################### Generate output in ClustalW input format above



    clustered_proteins_string = ["(%s and (%s))" % (p1, str(ResidueCompactor(p1, residues))) for (p1, residues) in clustered_proteins]
    extracted_template =  " or ".join(clustered_proteins_string)
    print "\nThe full extracted 3D template selection:", extracted_template
    bucket = {}
    for (p,res_list) in clustered_proteins:
        p1 = bucket.setdefault(p, {})
        for (s,c,r) in res_list:
            s1 = p1.setdefault(s, {})
            c1 = s1.setdefault(c, {})
            c1[(s,c,r)]=(s,c,r)
    # print "\nThe extracted 3D templates (ordered by each protein's sequence.  Therefore don't expect these to correspond.).  HET residues are represented by '?':"
    # for p in bucket: 
    #     for s in bucket[p]:
    #         for c in bucket[p][s]:
    #             res = bucket[p][s][c].keys()
    #             res.sort(keynat.keynat_cmp)
    #             print ("%s/%s/%s" % (p,s,c)).ljust(15), ": ", "".join([get_residue_name(residue_names_by_protein[p][r]) for r in res])
    print ""
cmd.extend("ligalign_find_template", find_template)

# color by distance to nearest residue
def color_by_closeness(baseline, active_site, residue_atoms, distances):
    if not pymol.stored.__dict__.has_key("ligalign_settings"):
        pymol.stored.ligalign_settings = {}
    if not pymol.stored.ligalign_settings.has_key("distant_residue_distance"):
        pymol.stored.ligalign_settings["distant_residue_distance"] = 4.5
    if not pymol.stored.__dict__.has_key("ligalign_colors") or not pymol.stored.ligalign_colors:
        pymol.stored.ligalign_colors = []
        for i in range(10):
            name = "ligalign_color%d"%i
            cmd.set_color(name, [(0.1*i), 1.0-(0.1*i), 0.0])
            pymol.stored.ligalign_colors.append(name)
    def find_color_bucket(distance):
        max_buckets = len(pymol.stored.ligalign_colors)
        color = None
        for i in range(0, max_buckets): #max_buckets-1, 0, -1):
            max_dist = float(pymol.stored.ligalign_settings["distant_residue_distance"])
#            print "dist:", distance, "i", i, "bucket:", i * (max_dist / max_buckets)
            if distance < i * (max_dist / max_buckets):
                color = pymol.stored.ligalign_colors[i]
                break
        if not color:
            color = pymol.stored.ligalign_colors[max_buckets-1]
        return color
    # foreach residue in active_site, color by min-centroid-distance  USEDTOBE: for each atom in centroid, color by min-centroid-distance
    for (residue, (res_name, atoms)) in residue_atoms.items():
        color = find_color_bucket(distances[residue])
#        print "active_site:%s residue:%s color:%s distance:%s"%(active_site, residue, color, distances[residue])
        (segi, chain, resi) = residue
        if len(segi) == 0:
            segi = "''"            
        if len(chain) == 0: 
            chain = "''"

        cmd.color(color, "(%s and segi %s and chain %s and resi %s)"%(active_site, segi, chain, resi))
# TODO: is this better?        cmd.color(color, "(%s_surface and segi %s and chain %s and resi %s)"%(active_site, segi, chain, resi))

def ligalign_color(baseline, active_site):
    baseline_atoms, baseline_res_names, baseline_centroids = compute_centroids(baseline)
    residue_atoms, res_names, protein_centroids = compute_centroids(active_site)
    dists = {}
    for (residue, nearby_residues) in find_distances(baseline_centroids, protein_centroids).items():
        (distance, nearest) = nearby_residues[0]
        dists[residue] = distance
    color_by_closeness(baseline, active_site,  residue_atoms, dists)
cmd.extend("ligalign_color", ligalign_color)


def get_similar_ligands(ligand, num_to_return=3):
    if not pymol.stored.__dict__.has_key("ligand_similarities") or not pymol.stored.ligand_similarities:
        print "loading similarity database..."
        pymol.stored.ligand_similarities = {}
        for line in file(os.getcwd()+"/similarityDB/ligSimilarity_0.85.pdbIDlist", "r"):  # Figure out where to find this file, robustly.
            words = line.split()
            if len(words) != 3: continue
            similars = pymol.stored.ligand_similarities.setdefault(words[0].lower(), []) # lowercase everything, so we can find similar ligands no matter the capitalization.
            similars.append(words[1])
            similars = pymol.stored.ligand_similarities.setdefault(words[1].lower(), [])
            similars.append(words[0])
    if ligand:
        similars = pymol.stored.ligand_similarities.get(ligand.lower(), [])
        return similars[0:num_to_return]
    else:
        return []

def fetch_pdbs(names):
    if not pymol.stored.__dict__.has_key("protein_names"):
        pymol.stored.protein_names = []    
    for p in names:
        error = True
        print "attempting to load \"",p,"\" locally..."
        # need to create a good object name from the filename
        object_name = os.path.splitext(os.path.basename(p))[0]
        object_name = cmd.safe_oname_re.sub("_", object_name)
        try:
            if len(object_name) > 0: # make sure we didn't throw everything away
                cmd.load(p, object_name)
                p = object_name # put the right object name into the pymol.stored.protein_names list
                error = False
        except CmdException:
            print "attempting to fetch \"",p,"\" from PDB..."
            error = cmd.fetch(p, object_name, async=0)
        if not error:
            if object_name not in pymol.stored.protein_names:
                pymol.stored.protein_names.append(object_name)
    cmd.deselect()

def find_ligands():
    # TODO: maybe test subgraph isomorphism as a means to detect substructures that match, within larger structures that don't?  i.e., ATP and FAD have adenine moiety.
    contenders = {}
    sorted = {}
    hack = {}
    for p in pymol.stored.protein_names:
        # pull out the hetatm residues in this protein
        pymol.stored.resi = {}
        cmd.select("ligalign_internal_temp", "(%s and (hetatm or het) and not resn HOH and not hydro)" % p)
        cmd.iterate("ligalign_internal_temp", "stored.resi[(resi,chain,alt)]=(resi,chain,alt)")
        for (resi,chain,alt) in pymol.stored.resi.keys():
            select = "%s and (hetatm or het)" % p
            if chain!='':
                select +=  (" and chain %s" % chain)
            if alt!='':
                select +=  (" and alt %s"   % alt)
            select += (" and resi %s and not hydro" % resi)
            cmd.select("ligalign_internal_temp", select)
            nAtoms = cmd.count_atoms("ligalign_internal_temp")
#            print resi, chain, select, nAtoms

            # now filter out the ones that are too tiny for us to do much analysis on (<=5 atoms)
            if nAtoms <= 5: continue
            print "analyzing", select
            m = cmd.get_model("ligalign_internal_temp")
            fingerprint = build_fingerprint(m)
            mol_fingerprint = [0] * len(fingerprint[0])
            for f in fingerprint:
                mol_fingerprint = [(x+y) for (x,y) in zip(mol_fingerprint, f)]
            mol_fingerprint = [x/nAtoms for x in mol_fingerprint]
            res_list = contenders.setdefault(p, [])
            res_list.append(((resi,chain,alt), nAtoms, mol_fingerprint))
        if p not in contenders:
            raise Exception("No suitable ligand in %s!  Maybe all of the HETATMs are less than six heavy atoms?" % p)
    cmd.delete("ligalign_internal_temp")


    original_query = pymol.stored.protein_names[0] # assumes the order of finding proteins has been preserved!
    if not original_query in contenders:
        raise QuietException("No suitable ligand in %s!" % original_query)
    differences = {}
    # for each ligand in the original query,
    for (query_residue, query_nAtoms, query_fingerprint) in contenders[original_query]:
        # build a list of closest-sized ligands in the other proteins
        differences[query_residue] = []
        for p in contenders: #need to add query-molecule in somewhere.  pymol.stored.protein_names[1:]:# don't get points for matching against other ligands in query molecule.  #
            ligand_list = contenders[p]
            diff_list = [((score_fingerprints(fingerprint,query_fingerprint)),abs(nAtoms-query_nAtoms), -nAtoms, (p,resi,chain,alt)) for ((resi,chain,alt),nAtoms,fingerprint) in ligand_list]
            diff_list.sort()
            differences[query_residue].append(diff_list[0]) # smallest difference

    #print "contenders:", contenders
    #print "differences:", differences

    # choose ligands w/ min error, breaking ties to larger ligands 
    score = []
    for residue in differences:
        # compute total error
        sum_error = 0
        sum_size_diff = 0
        sum_size = 0
        residues = []
        for (diff, size_diff, size, lig) in differences[residue]:
            sum_error += diff
            sum_size_diff  += size_diff
            sum_size  += size # remember, sizes are negative in differences and we want them to stay negative so that sort breaks ties to bigger ligands
            residues.append(lig)
        score.append((sum_error, sum_size_diff, sum_size, residues))
    score.sort()
    (min_error, min_size_diff, max_size, ligands) = score[0]
    #print "score:", score

    ligand_names = []
    # NEED LIGAND ORDER TO MATCH PROTEIN ORDER!
    for pname in pymol.stored.protein_names:
        for (p, resi, chain, alt) in ligands:
            if p==pname:
                ligand_name = p+"_ligand"
                ligand_names.append(ligand_name)
                select = "%s and (hetatm or het)" % p
                if chain!='':
                    select +=  (" and chain %s" % chain)
                if alt!='':
                    select +=  (" and alt %s"   % alt)
                select += (" and resi %s and not hydro" % resi)

                cmd.select(ligand_name, select)
    cmd.deselect()
    pymol.stored.ligand_names = ligand_names
    return ligand_names

def random_ligand():
    get_similar_ligands(None) # init ligand_similarities db
    return random.choice(pymol.stored.ligand_similarities.keys())

def ligalign(fragment_based_alignment, ligand_tuple):
    reset_stored()
    ligands = list(ligand_tuple)
    if len(ligands) == 0:
        ligands = [random_ligand()]
    if len(ligands) == 1:
        print "finding similar ligands..."
        ligands.extend(get_similar_ligands(ligands[0]))
    print "fetching PDB files..."
    fetch_pdbs(ligands)
    if len(pymol.stored.protein_names) < 1:
        raise Exception("No PDB files found for '%s'" % (", ".join(ligands)))
    print "extracting ligands..."
    ln = find_ligands()[:]
    # add junk mappings
    lig_and_maps = [(x,None) for x in ln]
    if fragment_based_alignment:
        print "fracturing ligands..."
        lig_and_maps = fracture_ligands(lig_and_maps[0], lig_and_maps[1:len(lig_and_maps)])
    print "aligning..."
    ln = align_by_ligand(lig_and_maps[0], lig_and_maps[1:len(lig_and_maps)])
    cmd.deselect()

    print "displaying..."
    ligand_union = build_ligand_union([x for (x,y) in ln])
    active_site  = build_active_site([x for (x,y) in ln])
#    viz(setting="7", full_residues="True") 
    viz(setting="11", full_residues="True") 
    cmd.deselect()
cmd.extend("ligalign",lambda *x,**y: ligalign(False, list(x)))
cmd.extend("ligalign_fragments",lambda *x,**y: ligalign(True, list(x)))

def reset_stored():
    # TODO: delete selections?
    cmd.reinitialize()
    pymol.stored.protein_names = []
    pymol.stored.ligand_names  = []
    pymol.stored.protein_to_fragment_proteins_map  = {}
    pymol.stored.best_hinges_cache = {}
    pymol.stored.ligalign_colors = []
    pymol.stored.molecule_mappings = {}    
    pymol.stored.ligand_mappings = {}    
cmd.extend("ligalign_reset_stored",reset_stored)

def get_pdbs(ligand):
    fetch_pdbs([ligand])
    sims = get_similar_ligands(ligand)
    fetch_pdbs(sims)

class Bondset:
    def __init__(self, from_m, to_m, mapp, c={}, f={}, o=None, hscache = {}, ancache = {}, clump=None ):
        self.model    = from_m
        self.closed   = c.copy() # PyMOL distributed with python < 2.3!  Use hashtables to implement sets
        self.front = f.copy()
        self.mapping  = mapp  # remember, mapping is in terms of ids
        self.target_model = to_m
        if o:
            self.open = o.copy()
        else:
            self.open = dict([(a.index,True) for a in from_m.atom if a.index not in c])
        self.hinge_score_cache = hscache
        self.atom_neighbors_cache = ancache
        if clump:
            self.clumps = clump
        else:
            self.clumps = self.calculate_clumps()
    def find_cycles(self):
        if len(self.model.atom) == 0:
            return {}
        arbitrary_atom = self.model.atom[0].index
        return self.find_cycles_helper(arbitrary_atom, None, [])
    def find_cycles_helper(self, atom, parent, stack):
        cycles = []
        if atom in stack: # we've found a cycle!
            cycle = stack[stack.index(atom):]
            cycles.append(cycle)
        else:
            # Recurse!
            stack.append(atom)
            children = self.atom_neighbors(atom)
            for child in children:
                if child != parent:
                    cycles.extend(self.find_cycles_helper(child, atom, stack))
            stack.pop()
        return cycles
    def calculate_clumps(self):
        # depth-first search to find cycles
        cycles = self.find_cycles()
        # union together cycles into cyclical systems
        while True:
            break_out  = False
            all_merged = True
            # cycles is a list of lists of atoms 
            for cycle_index in range(len(cycles)): # for each cycle
                for atom in cycles[cycle_index]:   # for each atom in each cycle, 
                    for cycle2 in range(cycle_index+1, len(cycles)):  # check that it's not in some other cycle
                        if atom in cycles[cycle2]:  # if it is, we need to merge
                            # need to merge cycles
                            all_merged = False
                            break_out = True
                            merged_cycle = cycles[cycle_index]
                            merged_cycle.extend([a for a in cycles[cycle2] if a not in merged_cycle])
                            del cycles[cycle2]
                            del cycles[cycle_index]
                            cycles.append(merged_cycle) # and insert the merged-cycle back into our list of cycles
                            break
                    if break_out: break
                if break_out: break
            if all_merged: break

        for cycle in cycles:
            cycle.sort()
        clumps = dict([(atom, cycle) for cycle in cycles for atom in cycle])
        return clumps
    def atom_neighbors(self, atom):
        if atom not in self.atom_neighbors_cache:
            atom_index = [a.index for a in self.model.atom].index(atom)
            neighbors = []
            # unforunately, bonds only record indexes, not ids
            for b in self.model.bond:
                if b.index[0] == atom_index:
                    neighbors.append(b.index[1])
                elif b.index[1] == atom_index:
                    neighbors.append(b.index[0])
#            return [self.model.atom[n].index for n in neighbors]
            self.atom_neighbors_cache[atom]=[self.model.atom[n].index for n in neighbors]
        return self.atom_neighbors_cache[atom]
    def expansion_set(self, a):
        return self.clumps.get(a,[a])
    def expand_frontier(self, a):
        clumps = self.expansion_set(a)  # if atom isn't part of a cycle, just add the atom.
        for atom in clumps:
            self.front.pop(atom, None) # using pop instead of del to suppress KeyError if atom not in open
            self.open.pop(atom, None) # using pop instead of del to suppress KeyError if atom not in open
            self.closed[atom] = True 
        # add it's neighbors (which haven't already been closed)
            for n in self.atom_neighbors(atom):
                if n not in self.closed:
                    self.front[n] = True 
        return len(clumps)
    def copy(self):
        return Bondset(self.model, self.target_model, self.mapping, self.closed, self.front, self.open, self.hinge_score_cache, self.atom_neighbors_cache, self.clumps)
    def frontier(self):
        return self.front.keys()
    def remaining_atoms(self):
        return self.open.keys()
    def __str__(self):
        return "ATOMSET:[open="+str(self.open.keys())+", closed="+str(self.closed.keys())+", frontier="+str(self.front.keys())+", clumps="+str(self.clumps)+"]"
    def check_remainder_connected(self):
        test = self.copy()
        remainder = test.remaining_atoms()
        if len(remainder) == 0:
            return True
        # choose arbitrary atom
        arbitrary_atom = remainder[0]
        # clean out frontier, leaving an arbitrarily chosen atom
        test.front = {arbitrary_atom:True}
        # take transitive closure over remaining atoms (from arbitrary starting point)
        while len(test.front) > 0:
            (arbitrary_atom,junk) = test.front.popitem()
            test.expand_frontier(arbitrary_atom)
        # if every atom hasn't been closed, there was something unreachable from our starting point
        remainder = test.remaining_atoms() 
#        print "  ---  ---  remaining atoms:", remainder, "arbitrary:", arbitrary_atom
        return len(remainder) == 0
    def free(self):
        return self.remaining_atoms()
    def hinge_score(self, group):
        # caching
        group.sort()
        key = tuple(group)
        if key not in self.hinge_score_cache: # or True:
            self.hinge_score_cache[key] = self.hinge_score_helper(group)
        return self.hinge_score_cache[key]
    def hinge_score_helper(self, group):
        rmsd = rmsd_on_models(self.model, self.target_model, self.mapping, group)
        return rmsd ** 2.0
    def compute_model_key(self):
        atoms1 = [a.id for a in self.model.atom]
        atoms2 = [a.id for a in self.target_model.atom]
        atoms1.sort()
        atoms2.sort()
        mapping_list = self.mapping.items()
        mapping_list.sort()
        return (tuple(atoms1),tuple(atoms2),tuple(mapping_list))
    def seeds(self):
        # return each leaf atom and one atom from each ring system
        
        # grab first atoms out of each ring system
        # there might be some redundancies, so do the usual wash through a set
        ring_atoms = dict([(c[0], True) for c in self.clumps.values() if c[0] not in self.closed]).keys()
        
        # find leaf atoms (atoms which only have one neighbor that is still open (or are alone, i.e. last atom standing))
        leaf_atoms = [a for a in self.free() if len([x for x in self.atom_neighbors(a) if x not in self.closed]) < 2]

        seed_list = list(ring_atoms)
        seed_list.extend(leaf_atoms)
#        print "seeds:", seed_list, "size:",len(seed_list), " rings=",ring_atoms," leafs=",leaf_atoms
        return seed_list

def find_reasonable_splits(atomset):
    global debug_call_count
    cutoff = pymol.stored.ligalign_settings["cutoff"]
    max_splits = pymol.stored.ligalign_settings["max_splits"]
    splits = 0
    pymol.stored.best_hinges_cache = {}
    print "Attempting to use",splits,"hinges on",len(atomset.model.atom),"atoms..."
    score, size, groups = best_hinges(atomset,splits)
    score = score ** 0.5
    print "... tested",debug_call_count,"fragmentations.  Best gave an RMSD=",score,"with groups=",groups
    debug_call_count=0
    while True:
        splits += 1
        print "Attempting to use",splits,"hinges... ",
        c, s, g = best_hinges(atomset,splits)
        c = c ** 0.5
        print ""
        print "... tested",debug_call_count,"fragmentations.  Best gave an RMSD=",c,"with groups=",g,"size=",s
        assert len(atomset.model.atom) == s
        debug_call_count=0
        diff = (score - c)
        if diff < score * cutoff :
            print "Fragmenting into",splits,"pieces gave less than a", cutoff*100,"% improvement.  Let's use the fewer groups:  RMSD=",score,"with groups=",groups
            break
        elif max_splits and splits >= max_splits:
            print "Let's only use",splits,"fragments."
            break
        else:
            score, size, groups = c, s, g
    return score, groups

def fracture_ligands((baseline, junk_mapping), ligands):
    if not pymol.stored.__dict__.has_key("ligand_mappings"):
        pymol.stored.ligand_mappings = {}    # cache the mappings for possible future display
    prefractured_mappings = []
    new_ligands = [baseline]
    new_ligands.extend(ligands)
    return_vals = [(baseline, junk_mapping)]
    baseline_model = cmd.get_model(baseline)
    for (ligand, mapping) in ligands:
        ligand_model = cmd.get_model(ligand)
        print "Building mapping for",ligand,"and",baseline

        if False: #True: # if we're using iterative mapping
            ############## Bootstrap first iteration! ##################
            #if not mapping:
            #    mapping = build_mapping(ligand, baseline)
            mapping = build_mapping(ligand, baseline)
            bondset = Bondset(ligand_model, baseline_model, mapping)
            score, groups = find_reasonable_splits(bondset)
            nAtoms = cmd.count_atoms(ligand)
            nAtoms2 = sum([len(g) for g in groups],0)
            print "size of original ligand (",ligand,")=",nAtoms," and size of groups=",nAtoms2
            assert nAtoms == nAtoms2

            ############# Now refine atom-mapping given initial fragmentation ###############
            refined_mapping = {}
            for g in groups:
                # ligand_fragment is a selection consisting of only atoms in g
                ligand_fragment = "(%s and (%s))" % (ligand, " or ".join(["(index %d)" % (index) for index in g]))
                # baseline_fragment is a selection consisting of only atoms in the baseline that correspond to atoms in g
                # note Mappings have IDs!
                baseline_fragment = "(%s and (%s))" % (baseline, " or ".join(["(id %d)" % (mapping[[a for a in ligand_model.atom if a.index == index][0].id]) for index in g]))
                # build a fragment-only mapping
                fragment_mapping = build_mapping(ligand_fragment, baseline_fragment, ligand, baseline)
                print "fragment_mapping:", fragment_mapping
                print "ligand_fragment:", ligand_fragment
                print "baseline_fragment:", baseline_fragment
                for i in fragment_mapping:
                    assert i not in refined_mapping
                refined_mapping.update(fragment_mapping)
#                print "refmap:", refined_mapping
            # extract out fragment's alignment
            for i in mapping:
                if mapping[i] != refined_mapping[i]:
                    print "DIFF: ", i, mapping[i], refined_mapping[i]
            
#            pymol.stored.best_hinges_cache = {}   
            mapping = refined_mapping
            pymol.stored.molecule_mappings[(ligand, baseline)] = refined_mapping

        # run real fragmentation
        if not mapping:
            mapping = build_mapping(ligand, baseline)
        bondset = Bondset(ligand_model, baseline_model, mapping)
        bondset.hinge_score_cache = {} # TODO: BUG: FIXME: How does the new Bondset have a full cache?  weird.
        bondset.atom_neighbors_cache = {} # TODO: BUG: FIXME: How does the new Bondset have a full cache?  weird.
        bondset.clumps = bondset.calculate_clumps() 
        assert len(bondset.hinge_score_cache) == 0
        score, groups = find_reasonable_splits(bondset)
        nAtoms = cmd.count_atoms(ligand)
        nAtoms2 = sum([len(g) for g in groups],0)
        print "size of original ligand (",ligand,")=",nAtoms," and size of groups=",nAtoms2, "\n"
        assert nAtoms == nAtoms2

        return_vals.append((ligand,mapping))
        fragment_id = 0
        protein_name = ligand[:-7] # chops '_ligand' off ligand name to regenerate protein name
        for g in groups:
            fragment_id += 1
            # create a duplicate ligand and protein for this group
            new_protein_name = "%s_%d" % (protein_name, fragment_id) 
            frag_name = "%s_ligand" % new_protein_name
            cmd.create(new_protein_name, protein_name)
            frag = "(%s and (%s))" % (new_protein_name, " or ".join(["(index %d)" % (index) for index in g]))
            cmd.select(frag_name, frag)
            # insert them in the ligand/protein lists, # TODO: in the correct spots
            pymol.stored.protein_names.append(new_protein_name)
            pymol.stored.ligand_names.append(frag_name)
            pymol.stored.protein_to_fragment_proteins_map.setdefault(protein_name, []).append(new_protein_name)
            new_ligands.append(frag_name)
            prefractured_mappings.append((frag_name,mapping))
            return_vals.append((frag_name,mapping))
        # TODO: delete the original ligand 
    cmd.deselect()
    for (name, mapping) in return_vals:
        pymol.stored.ligand_mappings[name] = (baseline, mapping)    # cache the mappings for possible future display
    return return_vals


debug_call_count = 0
debug_calls = 0
def best_hinges(atomset, splits, indent="", bound=sys.maxint):
    global debug_calls
    if not pymol.stored.__dict__.has_key("best_hinges_cache"):
        pymol.stored.best_hinges_cache = {}   
    # make hash-key
    free = atomset.free()
    free.sort()
    key = (splits, tuple(free), atomset.compute_model_key())
    # caching here
    if key not in pymol.stored.best_hinges_cache:
#        print ".... key:",key, " is not in cache.  Computing best hinges with bound=",bound
#        print "atomset:",atomset,"seeds:",atomset.seeds(),"splits:",splits,"bound:",bound
        answer = best_hinges_helper(atomset, splits, indent, bound)
        pymol.stored.best_hinges_cache[key] = answer
    return pymol.stored.best_hinges_cache[key] 
def best_hinges_helper(atomset, splits, indent, bound_score):
    global debug_call_count
    indent += "  "
    if bound_score < 0.0001:
        return None # someone's found an exact match, and we're just getting whacked by floating point precision
    elif splits == 0: 
        debug_call_count += 1 
        # no more splits, put all the remaining atoms into one group
        group = atomset.free()
        score = atomset.hinge_score(atomset.free())
        if score > bound_score:
            return None
        else:
            return score, len(atomset.free()), [atomset.free()]
    else:
        if len(atomset.free()) == 0: 
            group = atomset.free()
            debug_call_count += 1 
            # no more atoms, return an empty group
            return 0.0, 0, []
        else:
            # the option to use an extra hinge can only make the alignment better, not worse, so initialize the score to the score of one split less
            best = best_hinges(atomset, splits-1, indent, bound_score) # can't pass down bound b/c less splits on same atomset is harder!
#            print "1 atomset:",atomset,"seeds:",atomset.seeds(),"splits:",splits,"bound_score:",bound_score,"best:",best
            if not best: # can't beat bound
                best_score  = bound_score
                best_size   = None
                best_groups = None
            else:
                (best_score, best_size, best_groups) = best
                # a best could exist because it was calculated and cached at a looser bound,
                # which has since tightened, so we need to ensure that we can still use it.
                if bound_score <= best_score:
                    best_score  = bound_score
                    best_size   = None
                    best_groups = None
#            print ".... reasonable groups:", [x for x in reasonable_groups(atomset, indent)]
#            print "2 atomset:",atomset,"seeds:", atomset.seeds()
            for (group,rest) in reasonable_groups(atomset, indent): #, duplicates, work_reduction):
                group_score = atomset.hinge_score(group)
                group_size  = len(group)
                debug_call_count += 1 
#                print indent, "#:",debug_call_count,"MSD:",best_score,"group:",group," group_score=",group_score," size=",group_size ," and the rest is", rest, " rest.free:",rest.free(), " len(rest.free):",len(rest.free()), " with ",splits," total splits"
                print ("#%s"%debug_call_count), "",
                if ((group_score*group_size) / len(atomset.free())) >= best_score:
                    # we can tell the best_score will stay best, even if the error for 'rest' is 0
                    continue
                else:
                    candidate = best_hinges(rest, splits-1, indent, best_score)
                    if not candidate:
                        continue
                    (candidate_score,candidate_size,candidate_groups) = candidate
                    # combine
                    assert candidate_size == len(rest.free())
                    size  = (candidate_size + group_size)
#                    print indent, "candidate=",candidate
                    assert size == len(atomset.free())
                    score = ((candidate_score * candidate_size) + (group_score * group_size)) / size  #from unwrapping of RMSD
                    if score < best_score: # if scores are tied, use fewer hinges.  i.e, < not <= !
                        best_score  = score
                        best_size   = size
                        best_groups = list(candidate_groups)
                        best_groups.append(group)
            if best_score < bound_score:
                return (best_score, best_size, best_groups)
            else:
                return None
                    
def reasonable_groups(atomset, indent=""):
    min_fragment_size = int(pymol.stored.ligalign_settings["min_fragment_size"]) # small ligands tell us very little about the binding protein
    duplicates={}
    work_reduction={}
    for seed in atomset.seeds():
        seedgroup = list(atomset.expansion_set(seed))
        new_atomset = atomset.copy()
        new_atomset.front = {} # clear out old frontier
        new_atomset.expand_frontier(seed)
#        print "... ... ...", seed, new_atomset.check_remainder_connected()
        if new_atomset.check_remainder_connected():  # remainder still connected
            nAtoms =  len(atomset.free())
            max_part_size = math.ceil(nAtoms/2.0) # 20 -> 10, 21 -> 11
            for part_size in range(min_fragment_size,max_part_size):
                for (group, rest) in extend_to_size(seedgroup, new_atomset, part_size, part_size, work_reduction, indent):
#                    print "... ... ... ... ... ... ", group, rest, rest.free()
                    if len(rest.free()) >= min_fragment_size:
                        group.sort()
                        if tuple(group) not in duplicates:
                            duplicates[tuple(group)] = True
                            if rest.check_remainder_connected():
                                yield group,rest
                            else:
                                pass

# TODO: we can terminate fragment expansion early, if the number of disconnected components we've generated exceeds the size we have left to spend (if we have more size to spend, we could maybe still end up connecting them all).  Need to track the number of disconnected components but, since we take all rings in one piece, the number of disconnected components is equal to the size of the frontier!
def extend_to_size(seedgroup, atomset, size, max_size, work_reduction, indent=""):
    # make cache key: (size, frontier, group, atomset)
    front_dict = dict([(tuple(atomset.expansion_set(atom)),True) for atom in atomset.frontier()])
    front = front_dict.keys()
    front.sort()
    free = atomset.free()
    free.sort()
    group2 = list(seedgroup)
    group2.sort()
    #key = (size, max_size, tuple(front), tuple(free), tuple(group2), atomset.compute_model_key())
    key = (size, tuple(front), tuple(free), tuple(group2), atomset.compute_model_key())

    # caching here
    if key not in work_reduction:
        answer = extend_to_size_helper(seedgroup, atomset, size, max_size, work_reduction, indent)
        work_reduction[key] = answer
    return work_reduction[key]

def extend_to_size_helper(seedgroup, atomset, size, max_size, work_reduction, indent=""):
#    indent += "  "
    if size<=0:
        return [(seedgroup, atomset)]
    else:
        # remove duplicates
        unduplicated = {}
        for (group, rest) in extend_to_size(seedgroup, atomset, size-1, max_size, work_reduction, indent):
            group.sort()
            unduplicated[tuple(group)] = rest

        # grab rings?
        expanded_groups = []
        for (group_t, rest) in unduplicated.items():
            group = list(group_t)
            if len(group) > size:
                expanded_groups.append((group, rest))
            else:
                front_dict = dict([(tuple(rest.expansion_set(atom)),atom) for atom in rest.frontier()]) # remove duplicates
                for exp_set,atom in front_dict.items():
                    new_atomset = rest.copy()
                    new_atomset.expand_frontier(exp_set[0])
                    if True or (len(new_atomset.frontier()) <= max_size):
#                    if (len(new_atomset.frontier()) <= max_size):
                        expanded_groups.append((group+list(exp_set), new_atomset))
        return expanded_groups

# from Magnus Lie Hetland 
def floyd_warshall(weights, n):
    d = {-1: weights}
    for k in range(n):
        d[k] = {}
        for i in range(n):
            for j in range(n):
                if  d[k-1][i,k]==sys.maxint or d[k-1][k,j]==sys.maxint: 
                    second_sum = sys.maxint
                else: 
                    second_sum = d[k-1][i,k] + d[k-1][k,j] 
                d[k][i,j] = min(d[k-1][i,j], second_sum)            
    return d[n-1]

def weights_from_bonds(bonds,n):
    weights = {}
    # default values
    for i in range(n):
        weights[i,i] = 0
        for j in range(n):
            weights.setdefault((i,j), sys.maxint)
    # build symmetric connectivities
    for b in bonds:
        assert len(b.index) == 2
        i, j = b.index[0], b.index[1]
        weights[i,j] = 1
        weights[j,i] = 1
    return weights

def connectivity(model):
    nAtoms =  len(model.atom)
    weights = weights_from_bonds(model.bond, nAtoms)
    connections = floyd_warshall(weights, nAtoms)
    return connections

# assumes the model is connected
def build_fingerprint(model):
    nAtoms = len(model.atom)
    connections = connectivity(model)
    fingerprint = []
    counts = {}
    for x in range(nAtoms):
        counts[x] = {}
        for y in range(nAtoms):
            distance = connections[x,y]
            counts[x][distance] = counts[x].get(distance, 0)+1
        fingerprint.append([])
        for y in range(nAtoms): # can't be further apart than if the atoms were in a linear chain
            distance = counts[x].get(y, 0)
            if distance == 0:
                break
            fingerprint[x].append(distance)
    return fingerprint

def score_fingerprints(a,b):
    return pymol.stored.ligalign_settings["fingerprint_scoring"]((a,b))
    
def gcd(a, b):
    if b: return gcd(b, a % b)
    return a
def nlcm(nums): return reduce(lambda a, b: a * b / gcd(a, b), nums)

def arbitrary_scoring(a,b):
    # A variety of scoring functions are possible.  Here are a few.
    if False:
        # quadratically weighted diff
        edit_dist   = [abs(x-y) for (x,y) in zip(a,b)] # truncates to shorter list.  Is this sufficient or should we pad with 0?  Probably pad, but this is easier for now
        multipliers = range(len(edit_dist), 0, -1) 
        weighted    = [x*y*y for (x,y) in zip(edit_dist, multipliers)]
        return sum(weighted, 0)
    elif True:
        # linearly weighted diff
        edit_dist   = [abs(x-y) for (x,y) in zip(a,b)] # truncates to shorter list.  Is this correct or should we pad with 0?  Probably pad, but this is easier for now
        multipliers = range(len(edit_dist), 0, -1) 
        weighted    = [x*y for (x,y) in zip(edit_dist, multipliers)]
        return sum(weighted, 0)
    elif False:
        # heuristically weighted diff
        edit_dist   = [abs(x-y) for (x,y) in zip(a,b)] # truncates to shorter list.  Is this correct or should we pad with 0?  Probably pad, but this is easier for now
        multipliers = [4,4,4,3,3,3,2,2,2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1] # for sensitivity to further away atoms, extend this list.  Also, these weights are completely arbitrary.
        weighted    = [x*y for (x,y) in zip(edit_dist, multipliers)]
        return sum(weighted, 0)
    elif False:
        # exponentially weighted diff
        edit_dist = [abs(x-y) for (x,y) in zip(a,b)]#[0:4] # truncates to local area.  Is this correct or should we pad with 0?  Probably pad, but this is easier for now
        val = 0.0
        mult = 100000000000.0 # we'll get precision errors w/ very large ligands, the kind where differences in atoms connectivity get below floating point precision
        for x in edit_dist:
            val += mult * x
            mult *= 0.1
        return val
    elif False:
        # proportional to number of edits, but doesn't care how far away they are.
        edit_dist = [abs(x-y) for (x,y) in zip(a,b)][0:4] # truncates to local area.
        return sum(edit_dist,0)
    else:
        # hamming
        edit_dist = [1 for (x,y) in zip(a,b) if x!=y][0:4] # truncates to local area. 
        return sum(edit_dist,0)

def get_atom_id(model, atom_num):
    return model.atom[atom_num].id

def build_mapping(sel1, sel2, from_connectivity=None, to_connectivity=None):
    if not pymol.stored.__dict__.has_key("molecule_mappings"):
        pymol.stored.molecule_mappings = {}    
    # build key
    key = tuple([sel1, sel2, from_connectivity, to_connectivity])
    if key in pymol.stored.molecule_mappings:
        return pymol.stored.molecule_mappings[key]
    if not from_connectivity:
        from_connectivity = sel1
    if not to_connectivity:
        to_connectivity = sel2


    def score_maps(weights_set, model1, model2, group):
        nAtoms1 = len(model1.atom)
        nAtoms2 = len(model2.atom)

        ##### Now using Kuhn-Munkres
        # initialize matrix from tuples
        cost_matrix = [[0 for b in model2.atom] for a in model1.atom]
        for a,b in weights_set:
            cost_matrix[a][b] = weights_set[(a,b)]
        m = munkres.Munkres()
        indices = m.compute(cost_matrix)

        #convert to index from ids
#        print "indices:", indices, "len(model1):", len(model1.atom), "len(model2):", len(model2.atom)
        mapping = {}
        for a,b in indices:
            mapping[model1.atom[a].id] = model2.atom[b].id
        
        score = rmsd_on_models(model1, model2, mapping, group)
        return mapping, score

    model1  = cmd.get_model(sel1)
    model2  = cmd.get_model(sel2)
    # consider connectivity to entire ligand (to make the mapping), while only caring about alignment of fragment (for the scoring)!
    from_model = cmd.get_model(from_connectivity)
    to_model = cmd.get_model(to_connectivity)
    finger1 = build_fingerprint(from_model)
    finger2 = build_fingerprint(to_model)
    nAtoms1 = len(model1.atom)
    nAtoms2 = len(model2.atom)
    group = [a.index for a in model1.atom]

    # PyMOL's use of id and index seem inconsistent...
    convert_model1_index = {}
    for i in range(len(model1.atom)):
        for j in range(len(from_model.atom)):
            if model1.atom[i].id == from_model.atom[j].id:
                convert_model1_index[i] = j
    convert_model2_index = {}
    for i in range(len(model2.atom)):
        for j in range(len(to_model.atom)):
            if model2.atom[i].id == to_model.atom[j].id:
                convert_model2_index[i] = j

    weights = []
    foo_weights = {}
    for i in range(nAtoms1):
        for index2 in range(nAtoms2): 
            fingerprint1_index = convert_model1_index[i]
            fingerprint2_index = convert_model2_index[index2]
            weight = score_fingerprints(finger1[fingerprint1_index], finger2[fingerprint2_index])
            # include a bonus if the types of the atoms match.  Don't want this to be infinitely weighted, b/c we may want to align variants of ligands together
            if model1.atom[i].symbol == model2.atom[index2].symbol:
                weight += len(zip(finger1[fingerprint1_index], finger2[fingerprint2_index]))  # arbitrary, but chosen to be same as weight for distance 1
            
            i_weights = foo_weights.setdefault(i, {})
            i_nodes    = i_weights.setdefault(weight*10, [])
            i_nodes.append(index2)
            weights.append((i,index2,weight))

    # for each node i, for each node j in a set of nodes that all have equal weight,
    # build a weighting that 
    p_weights_list = [[]]
    if False:
        for i_prime in foo_weights:
            for i in foo_weights:
                foo_w_k = foo_weights[i].keys()
            foo_w_k.sort()
            foo_w_k.reverse()        
            most_important_weights = foo_w_k[0:1] # take the first (and second?) heaviest weighted 
            for weight in foo_weights[i]:
                weight_class = foo_weights[i][weight]
                wc = []
                if weight in most_important_weights and i==i_prime:
                    for j in weight_class:
                        w = []
                # make j the strongest in this weight class
                        for node in weight_class:
                            if node == j:
                                tup = (i, node, weight+1)
                            else:
                                tup = (i, node, weight)
                            w.append(tup)
                        wc.append(w)
                else:
                    wc.append([(i,node,weight) for node in weight_class])
                new_list = []
                for p in p_weights_list:
                    for w in wc:
                        l = list(p)
                        l.extend(w)
                        new_list.append(l)
            p_weights_list = new_list

    p_weights_list = [[]]
    weights_set = {}
    weights_to_check = {}
    for i in foo_weights:
        # first, initialize the weights
        for weight in foo_weights[i]:
            for node in foo_weights[i][weight]:
                weights_set[(i,node)] = weight

        # second, segregate the list into the ones where there's only a single maximally-weighted neighbor, and those where there's several
        foo_w_k = foo_weights[i].keys()
        foo_w_k.sort()
        foo_w_k.reverse()        
        weight_class = foo_weights[i][foo_w_k[0]]
        if len(weight_class) != 1:
            weights_to_check[(i, foo_w_k[0])] = weight_class


    best_mapping = None
    best_score = None
    decided_settings = []
    # TODO: BUG: FIXME:
    # We need to use two weights-loops so that we can break ties in two atom mappings at once.  Why is this useful?
    # Because a six-membered ring with para substituents has the two edges which look identical to the atom-graph-fingerprints!
    # That means that the ring could be flipped (rotated 180) and we need to break ties on *both* atoms on the side.
    for (i_1, weight_1), nodes_1 in weights_to_check.items():
        best_node_1 = None
        for node_1 in nodes_1:
            # make node the prefered partner (break ties in favor of node)
            weights_set[(i_1,node_1)] = weight_1+1
            (mapping,score) = score_maps(weights_set, model1, model2, group)
            if not best_score or score < best_score:
                best_score   = score
                best_node_1    = node_1
                best_mapping = mapping

            # reset it back
            weights_set[(i_1,node_1)] = weight_1

        # now fix the prefered partner to be the best found
        if best_node_1:
            decided_settings.append((i_1,best_node_1,weight_1+1))

    # Now, for the weights set favoring the best node for each tied mapping node, compute a final score
    for (a,b,c) in decided_settings:
        weights_set[(a,b)] = c
    (mapping,score) = score_maps(weights_set, model1, model2, group)
    if not best_score or score < best_score:
        best_mapping = mapping

    pymol.stored.molecule_mappings[key] = best_mapping
    return best_mapping


def debug(sel):
    if True:
        cmd.iterate(sel, 'print "name:%s resn:%s resi:%s chain:%s alt:%s elem:%s q:%s b:%s segi:%s id:%s idx:%s" % (name, resn, resi, chain, alt, elem, q, b, segi, ID, index)')
cmd.extend("ligalign_debug_selection", debug)



def ligalign_profile_harness():
    prof.runcall(find_template, 2.0, 1.0, "geometric",  "1pq2", "1w0g", "1bu7", "1h5z", "1n97")
    prof.runcall(find_template, 2.0, 1.0, "chemical",  "1pq2", "1w0g", "1bu7", "1h5z", "1n97")
    prof.runcall(find_template, 2.0, 0.8, "chemical",  "1pq2", "1w0g", "1bu7", "1h5z", "1n97")
    prof.runcall(find_template, 2.5, 0.8, "chemical",  "1pq2", "1w0g", "1bu7", "1h5z", "1n97")
    prof.runcall(find_template, 2.5, 0.8, "identity",  "1pq2", "1w0g", "1bu7", "1h5z", "1n97")

if False: #True: # set true to profile code
    cmd.do("ligalign 1pq2, 1w0g, 1bu7, 1h5z, 1n97")
    cmd.do("ligalign_viz full_residues=True, setting=9")
    import hotshot, hotshot.stats
    prof = hotshot.Profile("ligalign.prof")
#profile.run('cmd.do("ligalign_fragments 1gzf, 2qwx")')
#profile.run('ligalign(True, ("1gzf","2qwx"))')
#    prof.runcall(ligalign_profile_harness)
    ligalign_profile_harness()
    prof.close()
    stats = hotshot.stats.load("ligalign.prof")
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    stats.print_stats(50)

def grab_image():
    cmd.bg_color(color="gray80")
    cmd.bg_color(color="white")
    cmd.space("cmyk")
    cmd.set("antialias", 2)
    cmd.ray(2400, 2400)
    i=1
    while "pymol%d.png"%i in [x.split('/')[-1] for x in os.listdir(os.path.expanduser("~/Desktop"))]:
        i += 1
    cmd.png("~/Desktop/pymol%d.png"%i, dpi=300)
    pass
#cmd.extend("ligalign_grab_image", grab_image)

def demo1():
    cmd.do("ligalign 1xbb, 1y57")
    cmd.bg_color(color="grey80")
    cmd.bg_color(color="grey80")
    cmd.bg_color(color="white")
    cmd.bg_color(color="grey80")
    cmd.set("transparency", 0.0)
    cmd.color("green", "1xbb")    
    cmd.color("green", "1xbb_surface")
    cmd.color("purpleblue", "1y57")    
    cmd.color("purpleblue", "1y57_surface")
    cmd.color("yellow", "(ligand_union)")
    cmd.create("ligands", "(ligand_union)")
    cmd.show("sticks", "ligands")
    cmd.show("surface", "1xbb")    
    cmd.show("surface", "1y57")
    cmd.show("surface", "1xbb_surface")    
    cmd.show("surface", "1y57_surface")
    cmd.disable("1y57")
    cmd.disable("1y57_surface")
    cmd.clip("near", 500)
    cmd.clip("far", 500)
    cmd.zoom("1xbb", complete=1)
    cmd.rock()
def demo2():
    cmd.show("sticks", "1xbb_surface")
    cmd.show("sticks", "1y57_surface")
    cmd.enable("1y57_surface")
    cmd.hide("sticks", "1y57_surface and HET")
    cmd.hide("surface", "1xbb")
    cmd.hide("surface", "1y57")
    cmd.set("transparency", 0.8)
    cmd.zoom("1xbb_surface or 1y57_surface")
    cmd.origin("1xbb_surface or 1y57_surface")
cmd.extend("demo1", demo1)
cmd.extend("demo2", demo2)


print "LigAlign v1.00 loaded."
